@@ -51,6 +51,7 @@
 #include "runenvironment.h"
 #include <buildgraph/artifact.h>
 #include <buildgraph/buildgraph.h>
+#include <buildgraph/buildgraphloader.h>
 #include <buildgraph/emptydirectoriesremover.h>
 #include <buildgraph/nodetreedumper.h>
 #include <buildgraph/productbuilddata.h>
@@ -71,7 +72,7 @@
 #include <tools/installoptions.h>
 #include <tools/preferences.h>
 #include <tools/processresult.h>
-#include <tools/scannerpluginmanager.h>
+#include <tools/qbspluginmanager.h>
 #include <tools/scripttools.h>
 #include <tools/setupprojectparameters.h>
 #include <tools/qbsassert.h>
@@ -82,11 +83,8 @@
 #include <QtCore/qshareddata.h>
 
 #include <mutex>
-
-#ifdef QBS_STATIC_LIB
-extern "C" ScannerPlugin *cppScanners[];
-extern "C" ScannerPlugin *qtScanners[];
-#endif
+#include <utility>
+#include <vector>
 
 namespace qbs {
 namespace Internal {
@@ -111,11 +109,9 @@ static void loadPlugins(const QStringList &_pluginPaths, const Logger &logger)
             pluginPaths << pluginPath;
         }
     }
-    ScannerPluginManager::instance()->loadPlugins(pluginPaths, logger);
-#ifdef QBS_STATIC_LIB
-    ScannerPluginManager::instance()->loadPlugins(cppScanners);
-    ScannerPluginManager::instance()->loadPlugins(qtScanners);
-#endif
+    auto pluginManager = QbsPluginManager::instance();
+    pluginManager->loadStaticPlugins();
+    pluginManager->loadPlugins(pluginPaths, logger);
 
     qRegisterMetaType<ErrorInfo>("qbs::ErrorInfo");
     qRegisterMetaType<ProcessResult>("qbs::ProcessResult");
@@ -153,6 +149,7 @@ BuildJob *ProjectPrivate::buildProducts(const QList<ResolvedProductPtr> &product
 
     BuildJob * const job = new BuildJob(logger, jobOwner);
     job->build(internalProject, productsToBuild, options);
+    QBS_ASSERT(job->state() == AbstractJob::StateRunning,);
     return job;
 }
 
@@ -161,6 +158,7 @@ CleanJob *ProjectPrivate::cleanProducts(const QList<ResolvedProductPtr> &product
 {
     CleanJob * const job = new CleanJob(logger, jobOwner);
     job->clean(internalProject, products, options);
+    QBS_ASSERT(job->state() == AbstractJob::StateRunning,);
     return job;
 }
 
@@ -172,6 +170,7 @@ InstallJob *ProjectPrivate::installProducts(const QList<ResolvedProductPtr> &pro
         addDependencies(productsToInstall);
     InstallJob * const job = new InstallJob(logger, jobOwner);
     job->install(internalProject, productsToInstall, options);
+    QBS_ASSERT(job->state() == AbstractJob::StateRunning,);
     return job;
 }
 
@@ -203,14 +202,19 @@ QList<ResolvedProductPtr> ProjectPrivate::allEnabledInternalProducts(bool includ
     return enabledInternalProducts(internalProject, includingNonDefault);
 }
 
+static bool matches(const ProductData &product, const ResolvedProductConstPtr &rproduct)
+{
+    return product.name() == rproduct->name
+            && product.profile() == rproduct->profile
+            && product.multiplexConfigurationId() == rproduct->multiplexConfigurationId;
+}
+
 static ResolvedProductPtr internalProductForProject(const ResolvedProjectConstPtr &project,
                                                     const ProductData &product)
 {
     for (const ResolvedProductPtr &resolvedProduct : qAsConst(project->products)) {
-        if (product.name() == resolvedProduct->name
-                && product.profile() == resolvedProduct->profile) {
+        if (matches(product, resolvedProduct))
             return resolvedProduct;
-        }
     }
     for (const ResolvedProjectConstPtr &subProject : qAsConst(project->subProjects)) {
         const ResolvedProductPtr &p = internalProductForProject(subProject, product);
@@ -228,8 +232,11 @@ ResolvedProductPtr ProjectPrivate::internalProduct(const ProductData &product) c
 ProductData ProjectPrivate::findProductData(const ProductData &product) const
 {
     for (const ProductData &p : m_projectData.allProducts()) {
-        if (p.name() == product.name() && p.profile() == product.profile())
+        if (p.name() == product.name()
+                && p.profile() == product.profile()
+                && p.multiplexConfigurationId() == product.multiplexConfigurationId()) {
             return p;
+        }
     }
     return ProductData();
 }
@@ -725,14 +732,14 @@ RuleCommandList ProjectPrivate::ruleCommands(const ProductData &product,
                     case AbstractCommand::JavaScriptCommandType: {
                         externalCommand.d->type = RuleCommand::JavaScriptCommandType;
                         const JavaScriptCommandPtr &jsCmd
-                                = internalCommand.staticCast<JavaScriptCommand>();
+                                = std::static_pointer_cast<JavaScriptCommand>(internalCommand);
                         externalCommand.d->sourceCode = jsCmd->sourceCode();
                         break;
                     }
                     case AbstractCommand::ProcessCommandType: {
                         externalCommand.d->type = RuleCommand::ProcessCommandType;
                         const ProcessCommandPtr &procCmd
-                                = internalCommand.staticCast<ProcessCommand>();
+                                = std::static_pointer_cast<ProcessCommand>(internalCommand);
                         externalCommand.d->executable = procCmd->program();
                         externalCommand.d->arguments = procCmd->arguments();
                         externalCommand.d->workingDir = procCmd->workingDir();
@@ -756,6 +763,11 @@ static bool productIsRunnable(const ResolvedProductConstPtr &product)
     return product->fileTags.contains("application");
 }
 
+static bool productIsMultiplexed(const ResolvedProductConstPtr &product)
+{
+    return product->productProperties.value(QStringLiteral("multiplexed")).toBool();
+}
+
 void ProjectPrivate::retrieveProjectData(ProjectData &projectData,
                                          const ResolvedProjectConstPtr &internalProject)
 {
@@ -770,10 +782,12 @@ void ProjectPrivate::retrieveProjectData(ProjectData &projectData,
         product.d->version = resolvedProduct->productProperties
                                                         .value(QLatin1String("version")).toString();
         product.d->profile = resolvedProduct->profile;
+        product.d->multiplexConfigurationId = resolvedProduct->multiplexConfigurationId;
         product.d->location = resolvedProduct->location;
         product.d->buildDirectory = resolvedProduct->buildDirectory();
         product.d->isEnabled = resolvedProduct->enabled;
         product.d->isRunnable = productIsRunnable(resolvedProduct);
+        product.d->isMultiplexed = productIsMultiplexed(resolvedProduct);
         product.d->properties = resolvedProduct->productProperties;
         product.d->moduleProperties.d->m_map = resolvedProduct->moduleProperties;
         for (const GroupPtr &resolvedGroup : qAsConst(resolvedProduct->groups))
@@ -901,6 +915,7 @@ SetupProjectJob *Project::setupProject(const SetupProjectParameters &parameters,
     try {
         loadPlugins(parameters.pluginPaths(), logger);
         job->resolve(*this, parameters);
+        QBS_ASSERT(job->state() == AbstractJob::StateRunning,);
     } catch (const ErrorInfo &error) {
         // Throwing from here would complicate the API, so let's report the error the same way
         // as all others, via AbstractJob::error().
@@ -1120,6 +1135,44 @@ ErrorInfo Project::dumpNodesTree(QIODevice &outDevice, const QList<ProductData>
     return ErrorInfo();
 }
 
+Project::BuildGraphInfo Project::getBuildGraphInfo(const QString &bgFilePath,
+                                                   const QStringList &requestedProperties)
+{
+    BuildGraphInfo info;
+    try {
+        const Internal::TopLevelProjectConstPtr project = BuildGraphLoader::loadProject(bgFilePath);
+        info.bgFilePath = bgFilePath;
+        info.overriddenProperties = project->overriddenValues;
+        info.profileData = project->profileConfigs;
+        std::vector<std::pair<QString, QString>> props;
+        for (const QString &prop : requestedProperties) {
+            QStringList components = prop.split(QLatin1Char('.'));
+            const QString propName = components.takeLast();
+            props.push_back(std::make_pair(components.join(QLatin1Char('.')), propName));
+        }
+        for (const ResolvedProductConstPtr &product : project->allProducts()) {
+            if (props.empty())
+                break;
+            if (product->profile != project->profile())
+                continue;
+            for (auto it = props.begin(); it != props.end();) {
+                const QVariant value
+                        = product->moduleProperties->moduleProperty(it->first, it->second);
+                if (value.isValid()) {
+                    info.requestedProperties.insert(it->first + QLatin1Char('.') + it->second,
+                                                    value);
+                    it = props.erase(it);
+                } else {
+                    ++it;
+                }
+            }
+        }
+    } catch (const ErrorInfo &e) {
+        info.error = e;
+    }
+    return info;
+}
+
 #ifdef QBS_ENABLE_PROJECT_FILE_UPDATES
 /*!
  * \brief Adds a new empty group to the given product.
